-- proasic3/PA3_genlib.vhdl
-- created 20190618
-- version 20190809 (restart)
-- version : sam. août 10 08:06:58 CEST 2019
--
-- This file is just a little file that implements some gates and stuff from the
-- standard ProASIC3 cell library (Actel/Microsemi) in order
-- to simulate with a standalone tool (such as GHDL). See pa3_libguide_ug.pdf
--
-- It's used to check equivalence with files that start with these lines :
-- library proasic3;
--     use proasic3.all;
--
-- Since the original is made of proprietary code, I'm just rewriting the gates that I use,
-- along with a simple RTL implementation suitable for cycle-acurate simulations.
-- I added some delay but it is very coarse.

Library ieee;
    use ieee.std_logic_1164.all;

package PA3_genlib is

  subtype LookupType1  is std_logic_vector( 0 downto 0);
  subtype LookupType2  is std_logic_vector( 1 downto 0);
  subtype LookupType4  is std_logic_vector( 3 downto 0);
  subtype LookupType8  is std_logic_vector( 7 downto 0);

  type type_Histo8 is array (0 to 7) of integer;
  type type_Histo_entry is
    record
      index : integer;
      last  : integer;
      histo : type_Histo8;
    end record;
  type type_Histo_Array is array(integer range <>) of type_Histo_entry;
  type type_Histo_Access is access type_Histo_Array;
  shared variable HistoPtr : type_Histo_Access;

  shared variable gate_instance_counter : integer := 0;
  shared variable gate_select_number : integer := 0;
  shared variable bit_select_number : integer := -1;
  shared variable cycle_counter : integer := 0;

  type EnumGates2 is (
    AND2, AND2A, NAND2, NOR2A, NOR2B, NOR2,
    OR2A, OR2B, OR2, XNOR2, XOR2 );

  type EnumGates3 is (
     AND3,  AND3A,  AND3B,  AND3C, AO1,   AO12,  AO13,  AO14,  AO15,  AO16,
     AO17,  AO18,   AO1A,   AO1B,  AO1C,  AO1D,  AO1E,  AOI1,  AOI1A, AOI1B,
     AOI1C, AOI1D,  AOI5,   AX1,   AX1A,  AX1B,  AX1C,  AX1D,  AX1E,  AXO1,
     AXO2,  AXO3,   AXO5,   XO6,   AXO7,  AXOI1, AXOI2, AXOI3, AXOI4, AXOI5,
     AXOI7, MAJ3,   MX2,    MX2A,  MX2B,  MX2C,
     NAND3, NAND3A, NAND3B, NOR3,  NOR3A, NOR3B, NOR3C, OA1,
     OA1A,  OA1B,   OA1C,   OAI1,  OR3,   OR3A,  OR3B,  OR3C,  XA1,   XA1A,
     XA1B,  XA1C,   XO1,    XO1A,  XOR3,  XNOR3, ZOR3,  ZOR3I);

  subtype string9 is string(1 to 9);
  type string9x15 is array(1 to 15) of string9;
  constant EnumDFF : string9x15 := (
    "DFN1     ",
    "DFN1C0   ",
    "DFN1C1   ",
    "DFN1E0   ",
    "DFN1E0C0 ",
    "DFN1E0C1 ",
    "DFN1E0P0 ",
    "DFN1E0P1 ",
    "DFN1E1   ",
    "DFN1E1C0 ",
    "DFN1E1C1 ",
    "DFN1E1P0 ",
    "DFN1E1P1 ",
    "DFN1P0   ",
    "DFN1P1   ");

  function SLV_to_bin(s:std_logic_vector) return string;
  impure function Gate_Census(i : integer) return integer;
  function ComputeLUT4 (t: EnumGates2) return LookupType4;
  impure function LookItUp2 (A, B : std_logic; LUT4 : LookupType4; gate_number : integer) return std_logic;
  function LookItUp2_fast (A, B : std_logic; LUT4 : LookupType4) return std_logic;
  impure function InferLUT4(L: LookupType4; s: string; gate_number : integer) return LookupType4;
  function ComputeLUT8 (t: EnumGates3) return LookupType8;
  impure function LookItUp3 (A, B, C : std_logic; LUT8 : LookupType8; gate_number : integer) return std_logic;
  function LookItUp3_fast (A, B, C : std_logic; LUT8 : LookupType8) return std_logic;
  impure function InferLUT8(L: LookupType8; s: string; gate_number : integer) return LookupType8;
  procedure update_select_gate(gate_number : integer; bit_number : integer);
  procedure update_histogram;
  procedure display_histogram;

end PA3_genlib;

package body PA3_genlib is

  -- from YGREC8/YGREC8_utils.vhdl:
  function SLV_to_bin(s:std_logic_vector) return string is
    variable t : string(s'range);
    variable u : string(3 downto 1);
  begin
    for i in s'range loop
      u := std_logic'image(s(i));
      t(i) := u(2);
    end loop;
    return t;
  end SLV_to_bin;

  impure function Gate_Census(i : integer) return integer is
  begin
    gate_instance_counter := gate_instance_counter + 1;
    return gate_instance_counter;
  end Gate_Census;

  function ComputeLUT4 (t: EnumGates2) return LookupType4 is
    variable L : LookupType4;
    variable A, B, C, Y : std_logic;
    variable i : integer;
  begin
    i := 0;

    A := '0';
    loop

      B := '0';
      loop 

        case t is
          when AND2   => Y :=          A  and      B; 
          when AND2A  => Y :=     (not A) and      B ;
          when NAND2  => Y := not    ( A  and      B);
          when NOR2   => Y :=          A  nor      B ;
          when NOR2A  => Y := not((not A)  or      B) ;
          when NOR2B  => Y := not((not A)  or (not B));
          when OR2    => Y :=          A   or      B  ;
          when OR2A   => Y :=     (not A)  or      B  ;
          when OR2B   => Y :=     (not A)  or (not B) ;
          when XNOR2  => Y := not     (A  xor      B) ;
          when XOR2   => Y :=          A  xor      B  ;
          when others => report "Function not implemented : "
                  & EnumGates2'image(t) severity failure;
        end case;

        L(i) := Y;
        i := i+1;

        exit when B = '1'; 
        B := '1';
      end loop;

      exit when A = '1';
      A := '1';
    end loop;

    return L;
  end ComputeLUT4;

  impure function LookItUp2 (A, B : std_logic; LUT4 : LookupType4;
          gate_number : integer) return std_logic is
    variable index : integer := 0;
  begin
    if A = '1' then
      index := 2;
    else
      if A /= '0' then
        return a;
      end if;
    end if;

    if B = '1' then
      index := index+1;
    else
      if B /= '0' then
        return B;
      end if;
    end if;
    HistoPtr(gate_number).index := index;
    return LUT4(index);
  end LookItUp2;

  function LookItUp2_fast (A, B : std_logic; LUT4 : LookupType4) return std_logic is
    variable index : integer := 0;
  begin
    if A = '1' then
      index := 2;
    end if;
    if B = '1' then
      index := index+1;
    end if;
    return LUT4(index);
  end LookItUp2_fast;

  impure function InferLUT4(L: LookupType4;
        s: string; gate_number : integer) return LookupType4 is
    variable T : LookupType4 := L;
    variable i : integer;
  begin
    -- init the histogram:
    HistoPtr(gate_number).last := 3;
    for i in 0 to 3 loop
      HistoPtr(gate_number).histo(i) := 0;
    end loop;

    if gate_select_number < 1 then
      report " gate#" & integer'image(gate_number) & " : " & s;
    else
      if gate_select_number = gate_number
        and bit_select_number >= 0
        and bit_select_number < 4
       then
        T(bit_select_number) := not T(bit_select_number);
        report " LUT4: Flipping bit " & integer'image(bit_select_number)
             & " of gate #" & integer'image(gate_number);
      end if;
    end if;
    return T;
  end InferLUT4;


  function ComputeLUT8 (t: EnumGates3) return LookupType8 is
    variable L : LookupType8;
    variable A, B, C, Y : std_logic;
    variable i : integer;
  begin
    i := 0;

    A := '0';
    loop

      B := '0';
      loop 

        C := '0';
        loop

          case t is
            when AND3   => Y := (A and B) and C ;
            when AND3A  => Y := ((not A) and B) and C;
            when AND3B  => Y := ((not A) and (not B)) and C;
            when AND3C  => Y := ((not A) and (not B)) and (not C);
            when AO1    => Y := (A and B) or C;
            when AO12   => Y := ((not A) and B) or ((not A) and (not C)) or (B and (not C)) or (A and (not B) and C);
            when AO13   => Y := (A and B) or (A and (not C)) or (B and (not C));
            when AO14   => Y := (A and B) or (A and (not C)) or (B and (not C)) or ((not A) and (not B) and C);
            when AO15   => Y := (A and (not B) and C) or ((not A) and B and C) or ((not A) and (not B) and (not C));
            when AO16   => Y := (A and B and (not C)) or ((not A) and (not B) and C);
            when AO17   => Y := (A and B and C) or ((not A) and B and (not C)) or ((not A) and (not B) and C);
            when AO18   => Y := ((not A) and B) or ((not A) and (not C)) or (B and (not C));
            when AO1A   => Y := ((not A) and B) or C;
            when AO1B   => Y := (A and B) or (not C);
            when AO1C   => Y := ((not A) and B) or (not C);
            when AO1D   => Y := ((not A) and (not B)) or C;
            when AO1E   => Y := ((not A) and (not B)) or (not C);
            when AOI1   => Y := not ( (A and B) or C );
            when AOI1A  => Y := not ( ((not A) and B) or C );
            when AOI1B  => Y := not ( (A and B) or (not C) );
            when AOI1C  => Y := not ( ((not A) and (not B)) or C );
            when AOI1D  => Y := not ( ((not A) and (not B)) or (not C) );
            when AOI5   => Y := not ( ((not A) and B and C) or (A and (not B) and (not C)) );
            when AX1    => Y := ((not A) and B) xor C;
            when AX1A   => Y := not (((not A) and B) xor C);
            when AX1B   => Y := ((not A) and (not B)) xor C;
            when AX1C   => Y := (A and B) xor C;
            when AX1D   => Y := not(((not A) and (not B)) xor C);
            when AX1E   => Y := not((A and B) xor C);
            when AXO1   => Y := (A and B) or (B xor C);
            when AXO2   => Y := ((not A) and B) or (B xor C);
            when AXO3   => Y := (A and (not B)) or (B xor C);
            when AXO5   => Y := ((not A) and B) or ((not B) xor C);
            when XO6    => Y := (A and (not B)) or ((not B) xor C);
            when AXO7   => Y := ((not A) and (not B)) or (B xor C);
            when AXOI1  => Y := not ((A and B) or (B xor C));
            when AXOI2  => Y := not (((not A) and B) or (B xor C));
            when AXOI3  => Y := not ((A and (not B)) or (B xor C));
            when AXOI4  => Y := not ((A and B) or ((not B) xor C));
            when AXOI5  => Y := not (((not A) and B) or ((not B) xor C));
            when AXOI7  => Y := not (((not A) and (not B)) or (B xor C));
            when MAJ3   => Y := (A and B) or (A and C) or (B and C);
            when MX2    => if C='0' then  Y :=      A ; else Y :=      B ; end if;
            when MX2A   => if C='0' then  Y := (not A); else Y :=      B ; end if;
            when MX2B   => if C='0' then  Y :=      A ; else Y := (not B); end if;
            when MX2C   => if C='0' then  Y := (not A); else Y := (not B); end if;
            when NAND3  => Y := not ((A and B) and C);
            when NAND3A => Y := not (((not A) and B) and C);
            when NAND3B => Y := not (((not A) and (not B)) and C);
            when NOR3   => Y := not ((A or B) or C);
            when NOR3A  => Y := not ((not A) or B or C);
            when NOR3B  => Y := not ((not A) or (not B) or C);
            when NOR3C  => Y := not ((not A) or (not B) or (not C));
            when OA1    => Y := (A or B) and C;
            when OA1A   => Y := ((not A) or B) and C;
            when OA1B   => Y := (A or B) and (not C);
            when OA1C   => Y := ((not A) or B) and (not C);
            when OAI1   => Y := not ((A or B) and C);
            when OR3    => Y := (A or B) or C;
            when OR3A   => Y := ((not A) or B) or C;
            when OR3B   => Y := ((not A) or (not B)) or C;
            when OR3C   => Y := ((not A) or (not B) or (not C));
            when XA1    => Y := (A xor B) and C;
            when XA1A   => Y := (not (A xor B)) and C;
            when XA1B   => Y := (A xor B) and (not C);
            when XA1C   => Y := (not (A xor B)) and (not C);
            when XO1    => Y := (A xor B) or C;
            when XO1A   => Y := (not (A xor B)) or C;
            when XOR3   => Y := (A xor B) xor C;
            when XNOR3  => Y := not ((A xor B) xor C);
            when ZOR3   => Y := (A and B and C) or ((not A) and (not B) and (not C));
            when ZOR3I  => Y := not ((A and B and C) or ((not A) and (not B) and (not C)));
            when others => report "Function not implemented : "
                                & EnumGates3'image(t) severity failure;
          end case;

          L(i) := Y;
          i := i+1;

          exit when C = '1';
          C := '1';
        end loop;

        exit when B = '1'; 
        B := '1';
      end loop;

      exit when A = '1';
      A := '1';
    end loop;

    return L;
  end ComputeLUT8;

  impure function LookItUp3 (A, B, C : std_logic; LUT8 : LookupType8;
      gate_number : integer) return std_logic is
    variable index : integer := 0;
  begin
    if A = '1' then
      index := 4;
    else
      if A /= '0' then
        return a;
      end if;
    end if;

    if B = '1' then
      index := index+2;
    else
      if B /= '0' then
        return B;
      end if;
    end if;

    if C = '1' then
      index := index+1;
    else
      if C /= '0' then
        return C;
      end if;
    end if;

    HistoPtr(gate_number).index := index;
    return LUT8(index);
  end LookItUp3;

  function LookItUp3_fast (A, B, C : std_logic; LUT8 : LookupType8) return std_logic is
    variable index : integer := 0;
  begin
    if A = '1' then
      index := 4;
    end if;

    if B = '1' then
      index := index+2;
    end if;

    if C = '1' then
      index := index+1;
    end if;
    return LUT8(index);
  end LookItUp3_fast;


  impure function InferLUT8(L: LookupType8;
        s: string; gate_number : integer) return LookupType8 is
    variable T : LookupType8 :=	L;
  begin
    -- init the histogram:
    HistoPtr(gate_number).last := 7;
    for i in 0 to 7 loop
      HistoPtr(gate_number).histo(i) := 0;
    end loop;

    if gate_select_number < 1 then
      report " gate#" & integer'image(gate_number) & " : " & s;
    else
      if gate_select_number = gate_number
        and bit_select_number >= 0
        and bit_select_number <  8
       then
        T(bit_select_number) := not T(bit_select_number);
        report " Flipping bit " & integer'image(bit_select_number) 
             & " of gate #" & integer'image(gate_number);
      end if;
    end if;
    return T;
  end InferLUT8;

  procedure update_select_gate(
    gate_number : integer; bit_number : integer) is
    variable i, j : integer;
  begin
    report integer'image(gate_instance_counter) & " gates found.";
    if gate_number < 1 then
      report " gate number: " & integer'image(gate_number)
           & "  bit number: " & integer'image( bit_number);
      report " -- no gate function is altered --";
      report "List of the gates:";
    end if;
    gate_select_number := gate_number;
     bit_select_number :=  bit_number;
    HistoPtr := new type_Histo_Array(1 to gate_instance_counter);
  end update_select_gate;

  procedure update_histogram is
    variable i, j : integer;
  begin
    cycle_counter := cycle_counter + 1;
    for i in HistoPtr'range loop
      j := HistoPtr(i).index;
      HistoPtr(i).histo(j) := 
      HistoPtr(i).histo(j) + 1;
    end loop;
  end update_histogram;

  procedure display_histogram is
    variable i, j : integer;
  begin
    report "End of simulation ! " & integer'image(cycle_counter) & " cycles.";
    for i in HistoPtr'range loop
      if HistoPtr(i).last = 7 then
        report " h(" & integer'image(i) & ") = "
          & integer'image(HistoPtr(i).histo(0)) & "  "
          & integer'image(HistoPtr(i).histo(1)) & "  "
          & integer'image(HistoPtr(i).histo(2)) & "  "
          & integer'image(HistoPtr(i).histo(3)) & "  "
          & integer'image(HistoPtr(i).histo(4)) & "  "
          & integer'image(HistoPtr(i).histo(5)) & "  "
          & integer'image(HistoPtr(i).histo(6)) & "  "
          & integer'image(HistoPtr(i).histo(7));
      else -- HistoPtr(i).last = 3
        report " h(" & integer'image(i) & ") = "
          & integer'image(HistoPtr(i).histo(0)) & "  "
          & integer'image(HistoPtr(i).histo(1)) & "  "
          & integer'image(HistoPtr(i).histo(2)) & "  "
          & integer'image(HistoPtr(i).histo(3));
      end if;
    end loop;
  end display_histogram;

end PA3_genlib;
